{% load i18n %}

{% translate "Literal with a percent symbol at the end %" %}
{% translate "Literal with a percent % symbol in the middle" %}

{% translate "It is 100%" %}
{% translate "It is 100%" context "female" %}
{% translate "Looks like a str fmt spec %s but should not be interpreted as such" %}
{% translate "Looks like a str fmt spec % o but should not be interpreted as such" %}

{% translate "1 percent sign %, 2 percent signs %%, 3 percent signs %%%" %}
{% blocktranslate with name="Simon" %}{{name}} says: 1 percent sign %, 2 percent signs %%{% endblocktranslate %}
